@@ -44,11 +44,110 @@ deep_helpers = {
 
 
 
-def generate_config(devicename):
-    """Generate the configuration for the specified device and write it
+def render_file(name, env, output_filename=None, no_output=False,
+                no_squash_blanklines=False, debug=False, **render_vars):
+
+    """Read a configuration template file and render it.
+
+    The name of the file will be found in the supplied Jinja2
+    environment so cannot just be an absolute patht to a file.  '.j2'
+    will be appended to the name.
+
+    The render_vars will be passed to the Jinja2 render() method to
+    provide variables.  These will typically by the devicename, device
+    object and inventory but can be omitted when running a test
+    template (as it isn't a proper template but just one to test things
+    out in).
+
+    The return value is True if the template could be rendered and
+    False if not.
+    """
+
+
+    filename = name + ".j2"
+
+
+    # try to load the template using the supplied environment (which includes
+    # the directory search paths, so will be relative to those)
+
+    try:
+        template = env.get_template(filename)
+
+    except jinja2.exceptions.TemplateNotFound:
+        print("warning: template file not found in environment: %s - skipping"
+                  % filename,
+              file=sys.stderr)
+
+        return False
+
+
+    # render the template configuration
+
+    config = template.render(**render_vars)
+
+
+    # return (successfully), if output is disabled
+
+    if no_output:
+        if debug:
+            print("debug: output disabled", file=sys.stderr)
+
+        return True
+
+
+    # if not disabled, squash blank lines
+
+    if not no_squash_blanklines:
+        # remove any blank lines at the top
+        config = re.sub(r"^\n+", "", config)
+
+        # squash three or more blank lines into two (a single blank
+        # line)
+        config = re.sub(r"\n\n\n+", r"\n\n", config)
+
+        # the last line should end with a single newline
+        config = re.sub(r"\n\n+$", r"\n", config)
+
+
+    # write output to either standard output or a file, depending on
+    # the options specified
+
+    if output_filename:
+        output_filename_expanded = output_filename.replace("%", devicename)
+
+        if debug:
+            print("debug: writing to output file: %s"
+                      % output_filename_expanded,
+                  file=sys.stderr)
+
+        with open(output_filename_expanded, "w") as output_file:
+            print(config, file=output_file)
+
+    else:
+        if debug:
+            print("debug: writing to standard output", file=sys.stderr)
+
+        print(config)
+
+
+    # rendering succeeded
+
+    return True
+
+
+
+def render_config(devicename, inventory, env, output_filename,
+                  no_output=False, no_squash_blanklines=False,
+                  dump_device=False, debug=False):
+
+    """Render the configuration for the specified device and write it
     to either standard output or a file, depending on the command line
     options specified.
 
+    The configuration will be generated from a template file which is
+    found in the Jinja2 environment, with the filename based on the
+    platform and role.  This rendering is done by render_file().
+
     Returns True if the configuration is generated successfully.  If
     there is a problem of some sort which is not serious enough to
     abort the script (such as one device cannot be found), the function
@@ -57,9 +156,6 @@ def generate_config(devicename):
     """
 
 
-    # configuration
-
-
     if devicename not in inventory["devices"]:
         print("warning: device not found in inventory: %s - skipping"
                   % devicename,
@@ -110,105 +206,36 @@ def generate_config(devicename):
     platform = device["platform"]
 
 
-    # generate
-
-
     if debug:
         print("debug: generating configuration for: %s role: %s "
                   "platform: %s" % (devicename, role, platform),
               file=sys.stderr)
 
 
-    if not os.path.isdir(roles_dirname):
-        print("error: role directory does not exist: %s" % roles_dirname,
-              file=sys.stderr)
+    # work out the filename of the template for the platform and role
 
-        exit(1)
-
-    if not os.path.isdir(include_dirname):
-        print("error: include directory does not exist: %s" % include_dirname,
-              file=sys.stderr)
-
-        exit(1)
-
-
-    if not os.path.isdir(os.path.join(roles_dirname, platform)):
-        print("error: platform not found: %s used for device: %s"
-                  % (platform, devicename), file=sys.stderr)
-
-        exit(1)
-
-
-    # read the template file
-
-    role_filename = os.path.join(platform, role) + ".j2"
+    role_filename = os.path.join(platform, role)
 
     if debug:
         print("debug: using role file (relative to filesystem loader "
                   "directory): %s" % role_filename,
                file=sys.stderr)
 
-    try:
-        template = env.get_template(role_filename)
 
-    except jinja2.exceptions.TemplateNotFound:
-        print("error: role not found: %s for platform: %s" % (role, platform),
+    # render the template configuration file
+
+    if not render_file(role_filename, env, output_filename, no_output,
+                       no_squash_blanklines, devicename=devicename,
+                       device=device, inventory=inventory, **vars):
+
+        print("warning: failed to render template for: %s - skipping"
+                  % devicename,
               file=sys.stderr)
 
-        exit(1)
+        return False
 
 
-    # render the template
-
-    config = template.render(devicename=devicename, device=device,
-                             inventory=inventory, **vars)
-
-
-    # return, if output is disabled
-
-    if no_output:
-        if debug:
-            print("debug: output disabled", file=sys.stderr)
-
-        return True
-
-
-    # if not disabled, squash blank lines
-
-    if not no_squash_blanklines:
-        # remove any blank lines at the top
-        config = re.sub(r"^\n+", "", config)
-
-        # squash three or more blank lines into two (a single blank
-        # line)
-        config = re.sub(r"\n\n\n+", r"\n\n", config)
-
-        # the last line should end with a single newline
-        config = re.sub(r"\n\n+$", r"\n", config)
-
-
-    # write output to either standard output or a file, depending on
-    # the options specified
-
-    if output_filename:
-        output_filename_expanded = output_filename.replace("%", devicename)
-
-        if debug:
-            print("debug: writing to output file: %s"
-                      % output_filename_expanded,
-                  file=sys.stderr)
-
-        with open(output_filename_expanded, "w") as output_file:
-            print(config, file=output_file)
-
-    else:
-        if debug:
-            print("debug: writing to standard output", file=sys.stderr)
-
-        print(config)
-
-
-    # generation succeeded
+    # rendering succeeded
 
     return True
 
@@ -308,6 +335,14 @@ parser.add_argument(
     help="define variable for use in the template",
     metavar=("VAR", "VALUE"))
 
+parser.add_argument(
+    "-T", "--template",
+    dest="template_filename",
+    help="read the inventory, set up the environment and then render the "
+             "specified template file (from the roles or include "
+             "directories) then stop - used to quickly test template "
+             "functions without needing to construct a device")
+
 parser.add_argument(
     "-q", "--quiet",
     action="store_true",
@@ -353,6 +388,7 @@ list_env = args.list_env
 dump_inventory = args.dump_inventory
 dump_filepaths = args.dump_filepaths
 dump_device = args.dump_device
+template_filename = args.template_filename
 quiet = args.quiet
 devicenames = args.devicename
 debug = args.debug
@@ -375,6 +411,21 @@ debug: device names: %s"""
           file=sys.stderr)
 
 
+# check that the roles and include directories exist
+
+if not os.path.isdir(roles_dirname):
+    print("error: roles directory does not exist: %s" % roles_dirname,
+            file=sys.stderr)
+
+    exit(1)
+
+if not os.path.isdir(include_dirname):
+    print("error: include directory does not exist: %s" % include_dirname,
+            file=sys.stderr)
+
+    exit(1)
+
+
 # check a couple of nonsensical configurations aren't being use related
 # to multiple devices
 
@@ -387,7 +438,6 @@ if (len(devicenames) > 1) and (not (no_output or dump_device)):
 
         exit(1)
 
-
     elif output_filename.find("%") == -1:
         print("error: multiple device names specified but output "
                   "filename does not contain a '%' to substitute the "
@@ -479,10 +529,20 @@ if list_env:
 
 
 
-# --- generate ---
+# --- render ---
 
 
 
+# if a template file has been specified directly (for testing) then
+# render that directly and stop
+
+if template_filename:
+    render_file(template_filename, env, output_filename, no_output,
+                no_squash_blanklines, debug)
+
+    exit(0)
+
+
 # go through all the devices specified, generate and write out their
 # configurations
 
@@ -490,7 +550,7 @@ if not devicenames:
     print("warning: no device names specified", file=sys.stderr)
 
 
-# this flag will change to False if any configuration fails to generate
+# this flag will change to False if any configuration fails to render
 # and is used to affect the return code from the script
 
 complete_success = True
@@ -501,14 +561,18 @@ for devicename in devicenames:
         print(devicename)
 
     try:
-        complete_success &= generate_config(devicename)
+        complete_success &= render_config(devicename, inventory, env,
+                                          output_filename, no_output,
+                                          no_squash_blanklines, dump_device,
+                                          debug)
 
     except helpers.HelperBaseException as e:
         # one of the explicit helper exceptions has occurred
 
 
-        # if we're raising an exception anyway, do that; if not, just print
-        # the exception text in the form of a user-friendly error message
+        # if we're raising an exception anyway, do that; if not, just
+        # print the exception text in the form of a user-friendly error
+        # message
 
         if raise_exception:
             raise
